/*

	I n s t r u c t i o n    C o m p i l e r

	(c) 1992  Perex soft

*/

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <iostream.h>
#include <fstream.h>
#include <stdlib.h>

const
  maxLineSize = 128,
  maxFiles    = 10;

const char commandNames[][10] = {
"include", "macro", "endall", "cross", "copy", ""
};

char instr	[256];
char pref_ed 	[256];
char pref_dd	[256];
char pref_fd	[256];
char pref_cb	[256];
char pref_dd_cb	[256];
char pref_fd_cb	[256];

ofstream ofile;				// output stream
ifstream files[maxFiles];		// maximum 10 includes
char fileNames[maxFiles][79];		// nazvy souboru
int numberLine[maxFiles];
int openFiles;
char line[ maxLineSize ];
int pline;

void clearFlags()
{
  memset( instr, 0, 256 );
  memset( pref_ed, 0, 256 );
  memset( pref_dd, 0, 256 );
  memset( pref_fd, 0, 256 );
  memset( pref_cb, 0, 256 );
  memset( pref_dd_cb, 0, 256 );
  memset( pref_fd_cb, 0, 256 );
  instr[0xed] = instr[0xdd] = instr[0xfd] = instr[0xcb] = 1;
  pref_dd[0xcb] = pref_fd[0xcb] = 1;
}

void closeAll()
{
  int i = 0;
  while ( i <= openFiles ) files[i++]. close();
  ofile. close();
}

void error( int err )
{
  char *s;

  switch ( err ) {
    case 1: s = "Unexpected end of file."; break;
    case 2: s = "Unknown command."; break;
    case 3: s = "Error in write to output file."; break;
    case 4: s = "Unexpected end of copy."; break;
    case 5: s = "Invalid code of instruction."; break;
    case 6: s = "Instruction can not be redefined."; break;
  }
  cout << "Error: " << s;
  cout << "\nIn file: " << fileNames[ openFiles ] << "      Line: "
       << numberLine[openFiles] << "\n\n";
  closeAll();
  exit( 1 );
}

void openFile( const char s[] )
{
  files[ ++openFiles ]. open( s );
  if ( !files[ openFiles ] )
    {
      cout << "I can not open include file: " << s;
      exit( 1 );
    }
  strcpy( fileNames[ openFiles ], s );
  numberLine[ openFiles ] = 0;
}

void openOFile( const char s[] )
{
  ofile. open( s );
  if ( !ofile )
    {
      cout << "I can not open output file: " << s;
      exit( 1 );
    }
}

int getLine()
{
  files[ openFiles ]. getline( line, maxLineSize );
  numberLine[ openFiles ]++;
  int eof = files[ openFiles ]. eof();
  while ( eof && (openFiles > 0) )
    {
      files[ openFiles-- ]. close();
      eof = files[ openFiles ]. eof();
    };
  return eof;
}

void putLine( const char s[] )
{
  ofile. write( s, strlen(s) );
  if ( ofile. bad() ) error( 3 );
}

void crlf()
{
  putLine( "\n" );
}

void tab()
{
  putLine( "\t\t" );
}

int skipSpaces()
{
  char c = 0;
  while ( ((c = line[pline]) == ' ') || (c == 9) ) pline++;
  return line[pline] == 0;
}

int skipFreeLine()
{
  do {
    if ( getLine() ) return 1;
    pline = 0;
  } while ( skipSpaces() );
  int pos = pline;
  char c;
  while ( line[pos] != 0 )
    {
      while ( (*(int*)&line[pos] != '//') && (line[pos] != 0) ) pos++;
      if ( *(int*)&line[pos] ) line[pos] = 0;
    }
  pos = strlen(line);
  while ( (pos > 0) && ((c = line[pos]) == ' ') || (c == 9)) pos--;
  line[pos] = 0;
  if (pos == 0) return skipFreeLine();
    else return 0;
}

int isHexa( char c )
{
  return ( ((c >= '0') && (c <= '9')) || ((c >= 'a') && (c <= 'f')) );
}

int decodeCommand()
{
  int pos = 0;
  int end = 0;
  while ( (commandNames[pos][0] != 0) && !end )
    {
      int pos1 = 0;
      while ( (commandNames[pos][pos1] == line[pline + pos1]) &&
	      ( commandNames[pos][pos1] != 0 ) ) pos1++;
      end = commandNames[pos][pos1] == 0;
      if ( end )
	{  pline += pos1; skipSpaces(); }
      pos++;
    }
  if ( end ) return --pos;
    else
   {
     if ( isHexa( line[pline] ) && isHexa( line[pline + 1] ) )
       {
	 pline = 1;
	 return 0xff;
       }
     return -1;
   }
}

void copy()
{
  int end = 0;
  crlf();
  do {
    if ( skipFreeLine() ) error( 4 );
    end = line[0] == '#';
    if ( !end ) {  tab(); putLine( &line[pline] ); crlf(); };
  } while ( !end );
  crlf();
}

unsigned char hexa( char c )
{
  if ( !isHexa( c ) ) error( 5 );
  if ( (c >= '0') && ( c <= '9' ) ) return c - '0';
  return toupper(c) - 'A' + 10;
}

unsigned char getHexa()
{
  pline += 2;
  return ( hexa( line[pline-2] ) << 4 ) | hexa( line[pline - 1] );
}

void makeLabel( const char s[], int idx )
{
  char s1[20];
  sprintf(s1, "_%X", idx );
  putLine( "lbl" ); putLine( s ); putLine( s1 ); putLine( ":" );
  crlf();
}

void instruction()
{
  unsigned char idx = getHexa();
  switch ( idx ) {
    case 0xed:
      pline++; idx = getHexa();
      if ( pref_ed[idx] ) error( 6 );
      pref_ed[idx] = 1; makeLabel("_ED", idx);
      break;
    case 0xcb:
      pline++; idx = getHexa();
      if ( pref_cb[idx] ) error ( 6 );
      pref_cb[idx] = 1; makeLabel("_CB", idx);
      break;
    case 0xdd:
      pline++; idx = getHexa();
      if ( idx == 0xcb )
	{
	  pline++;
	  idx = getHexa();
	  if ( pref_dd_cb[idx] ) error( 6 );
	  pref_dd_cb[idx] = 1; makeLabel("_DD_CB", idx);
	}
       else
	{
	  if ( pref_dd[idx] ) error( 6 );
	  pref_dd[idx] = 1; makeLabel("_DD", idx);
	}
      break;
    case 0xfd:
      pline++; idx = getHexa();
      if ( idx == 0xcb )
	{
	  pline++;
	  idx = getHexa();
	  if ( pref_fd_cb[idx] ) error( 6 );
	  pref_fd_cb[idx] = 1; makeLabel("_FD_CB", idx);
	}
       else
	{
	  if ( pref_fd[idx] ) error( 6 );
	  pref_fd[idx] = 1; makeLabel("_FD", idx);
	}
      break;
    default:
      if ( instr[idx] ) error( 6 );
      instr[idx] = 1; makeLabel("", idx);
  }
  copy();
}

void macro()
{
  skipSpaces();
  int pos = pline;
  char s[30];
  char c;
  while ( (( c = line[pos] ) != ' ') && ( c != '\t' ) && ( c != 0 ) )
    s[(pos++)-pline] = c;
  s[pos - pline] = 0;
  if ( c != 0 ) pline = pos;
    else line[pline] = 0;
  putLine( s );
  putLine( "\t\tmacro" );
  putLine( &line[pline] ); crlf();
  copy();
  putLine( "endm\r\n" );
}

void cross1( const char s1[], const char s2[], const char ins[] )
{
  char s3[10];
  putLine( s1 ); crlf();
  for (int i = 0; i < 256; i++)
    {
      sprintf(s3, "_%X", i );
      tab(); putLine( "dw\toffset " );
      if ( ins[i] )
	{
	  putLine("lbl"); putLine(s2); putLine(s3);
	}
       else
	{
	  putLine("err_instr");
	}
      crlf();
    }
  crlf();
}

void cross()
{
  cross1("bas_instr:", "", instr);
  cross1("prefix_ed:", "_ED", pref_ed);
  cross1("prefix_cb:", "_CB", pref_cb);
  cross1("prefix_dd:", "_DD", pref_dd);
  cross1("prefix_fd:", "_FD", pref_fd);
  cross1("prefix_dd_cb:", "_DD_CB", pref_dd_cb);
  cross1("prefix_fd_cb:", "_FD_CB", pref_dd_cb);
}

void go()
{
  int end = 0;
  clearFlags();
  while ( !end )
    {
      if ( skipFreeLine() ) error( 1 );
      if ( *(int *)&line[pline] != '/*' )
	{
	  if ( line[pline++] != '#' ) error( 2 );
	  switch ( decodeCommand() ) {
	    case -1: error( 2 ); break;
	    case  0: openFile( &line[pline] ); break;
	    case  1: macro(); break;
	    case  2: end = 1; break;
	    case  3: cross(); break;
	    case  4: copy(); break;
	    case 0xff: instruction(); break;
	  }
	}
       else
	{
	  do {
	    if ( skipFreeLine() ) error( 2 );
	  } while ( *(int *)&line[pline] != '*/' );
	}
    }
  cout << "Compilation successfull ...\n\n";
}

void main(int argc, char *argv[] )
{
  openFiles = -1;
  if ( argc <= 1 ) openFile( "Z80.INS" );
    else openFile( argv[1] );
  if ( argc <= 2 ) openOFile( "Z80.ASM" );
    else openOFile( argv[2] );
  go();
  closeAll();
}
